from dataclasses import dataclass
from typing import List

from tvalmetrics.scorers.augmentation_accuracy_scorer import AugmentationAccuracyScorer
from tvalmetrics.scorers.retrieval_precision_scorer import RetrievalPrecisionScorer


@dataclass
class AugmentationPrecision:
    """Augmentation precision score and information used to calculate the score.

    Fields
    ------
    score: float
        Float between 0 and 1 representing the augmentation precision score.
    question: str
        The question that was asked.
    answer: str
        The answer that was generated by the RAG system.
    context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    context_relevant_list: List[bool]
        List of booleans representing whether each context in context_list is relevant
        for answering the question.
    answer_contains_context_list: List[bool]
        List of booleans representing whether information from each context in
        context_list is in answer.
    """

    score: float
    question: str
    answer: str
    context_list: List[str]
    context_relevant_list: List[bool]
    answer_contains_context_list: List[bool]


class AugmentationPrecisionScorer(object):
    """Class for calculating augmentation precision score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model
        self.augmentation_accuracy_scorer = AugmentationAccuracyScorer(self.model)
        self.retrieval_precision_scorer = RetrievalPrecisionScorer(self.model)

    def score(
        self, question: str, answer: str, context_list: List[str]
    ) -> AugmentationPrecision:
        """Calculate augmentation precision score.

        Parameters
        ----------
        question: str
            The question that was asked.
        answer: str
            The answer that was generated by the RAG system.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        AugmentationPrecisionScore
            Augmentation precision score and information used to calculate the score.
        """
        retrieval_precision_score = self.retrieval_precision_scorer.score(
            question, context_list
        )
        context_relevant_list = retrieval_precision_score.context_relevant_list
        augmentation_accuracy_score = self.augmentation_accuracy_scorer.score(
            answer, context_list
        )
        contains_context_list = augmentation_accuracy_score.answer_contains_context_list

        score = self.score_from_context_labels(
            context_relevant_list, contains_context_list
        )
        return AugmentationPrecision(
            score,
            answer,
            question,
            context_list,
            context_relevant_list,
            contains_context_list,
        )

    @staticmethod
    def score_from_context_labels(
        context_relevant_list: List[bool], answer_contains_context_list: List[bool]
    ) -> float:
        relevant_context_count = 0
        relevant_context_used_count = 0
        for context_relevant, contains_context in zip(
            context_relevant_list, answer_contains_context_list
        ):
            if context_relevant:
                relevant_context_count += 1
                if contains_context:
                    relevant_context_used_count += 1
        if relevant_context_count == 0:
            augmentation_precision = 0.0
        else:
            augmentation_precision = (
                relevant_context_used_count / relevant_context_count
            )

        return augmentation_precision



import logging
from typing import List

logger = logging.getLogger()


def parse_boolean_response(response: str) -> bool:
    """Parse boolean response from LLM evaluator.

    Attempts to parse response as true or false.

    Parameters
    ----------
    response: str
        Response from LLM evaluator.

    Returns
    -------
    bool
        Whether response should be interpreted as true or false.
    """
    response_lower = response.lower()
    if response_lower == "true":
        return True
    if response_lower == "false":
        return False
    logger.debug(f"Relevance response {response_lower} is not true or false")
    if "true" in response_lower and "false" not in response_lower:
        return True
    if "false" in response_lower and "true" not in response_lower:
        return False
    log_message = (
        f"Could not determine true or false from response {response_lower}"
        ", returning False"
    )
    logger.debug(log_message)
    return False


def parse_bullet_list_response(response: str) -> List[str]:
    """Parse bullet list response from LLM evaluator.

    Attempts to parse repsonse as a bullet list, returning a list of strings that
    correspond to the bullet points. The response is assumed to be a bullet list in
    markdown format with the bullet points denoted by asterisks.

    Parameters
    ----------
    response: str
        Response from LLM evaluator.

    Returns
    -------
    List[str]
        List of strings that correspond to the bullet points in the response.
    """
    if "*" not in response:
        log_message = (
            f"Response {response} does not contain bullet list. Returning all of "
            "response as main point."
        )
        logger.debug(log_message)
        return [response]
    if not response.startswith("*"):
        log_message = (
            f"Response {response} does not start with bullet, when it should be a "
            "bulleted list. Content before the first bullet will be removed."
        )
        logger.debug(log_message)
    bullet_list = response.split("*")[1:]
    bullet_list = [bullet.strip() for bullet in bullet_list]
    return bullet_list



import logging
import time
from typing import Dict, List, Optional, Tuple, Union

import openai
from openai.error import RateLimitError

logger = logging.getLogger()


def make_request_with_retry(
    messages: List[Dict[str, str]],
    model: str,
    temperature: float,
    wait_time: float = 0.1,
    max_retries: int = 12,
    exp_delay_base: int = 2,
) -> Optional[Dict[str, Union[List[Dict[str, Dict[str, str]]], Dict[str, int]]]]:
    """Makes request to OpenAI API with exponential back off retry logic.

    Parameters
    ----------
    messages: List[Dict[str, str]]
        List of messages to send to OpenAI API.
    model: str
        Name of the LLM model to use. Must be an Open AI chat completion model.
    temperature: float
        Float between 0 and 1. Temperature to use for Open AI model.
    wait_time: float
        Initial time to wait between retries.
    max_retries: int
        Maximum number of retries before stop retrying.
    exp_delay_base: int
        Base for exponential back off delay between retries.

    Returns
    -------
    Optional[Dict[str, Union[List[Dict[str, Dict[str, str]]], Dict[str, int]]]]
        Response from OpenAI API.
    """
    num_retries = 0
    while num_retries < max_retries:  # if it can't make the request - this is to retry
        try:
            completion = openai.ChatCompletion.create(
                model=model, messages=messages, temperature=temperature
            )
            return completion

        except RateLimitError:
            log_message = (
                "hit openai.error.RateLimitError and entered retry "
                f"logic, num_retries={num_retries}"
            )
            logger.debug(log_message)
            time.sleep(wait_time)
            wait_time *= exp_delay_base
            num_retries += 1
        except Exception as e:
            error_message = (
                "hit non openai.error.RateLimitError exception and "
                f"entered retry logic, num_retries={num_retries}, "
                f"exception {e}"
            )
            logger.error(error_message, exc_info=True)
            time.sleep(wait_time)
            wait_time *= exp_delay_base
            num_retries += 1

    logger.debug(f"reached max retries of {max_retries} for {model} request")
    return None


def get_message_response(
    open_ai_message_list: List[Dict[str, str]], model: str, temperature: float = 1.0
) -> Tuple[str, Dict[str, int]]:
    """Sends message list to Open AI API, parses response, and returns response message.

    Parameters
    ----------
    open_ai_message_list: List[Dict[str, str]]
        List of messages to send to OpenAI API.
    model: str
        Name of the LLM model to use. Must be an Open AI chat completion model.
    temperature: float
        Float between 0 and 1. Temperature to use for Open AI model.

    Returns
    -------
    Tuple[str, Dict[str, int]]
        Tuple of response message and usage dictionary.
    """
    completion = make_request_with_retry(open_ai_message_list, model, temperature)
    if completion is None:
        raise Exception(f"Failed to get message response from {model}, max retires hit")

    message_response = completion["choices"][0]["message"]["content"]  # type: ignore
    usage = completion["usage"]

    return message_response, usage  # type: ignore


def get_single_message(
    message: str,
    model: str,
    temperature: float,
    system_message: str = "You are a helpful assistant. Respond using markdown.",
) -> str:
    """Sets system message and sends message to Open AI API, returns response message.

    Parameters
    ----------
    message: str
        Message to send to OpenAI API.
    model: str
        Name of the LLM model to use. Must be an Open AI chat completion model.
    temperature: float
        Float between 0 and 1. Temperature to use for Open AI model.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    message_list = [
        {"role": "system", "content": system_message},
        {"role": "user", "content": message},
    ]

    try:
        response_message, _ = get_message_response(
            message_list, model, temperature=temperature
        )
    except Exception:
        error_message = (
            "An exception occurred while retrieving an answer from "
            f"{model}, using error message as response"
        )
        logging.error(error_message, exc_info=True)
        response_message = error_message

    return response_message



from dataclasses import dataclass
from statistics import mean
from typing import Dict, List, Optional

import pandas as pd
from tvalmetrics.scorers import (  # type: ignore
    AnswerConsistencyBinaryScorer,
    AnswerConsistencyScorer,
    AnswerSimilarityScorer,
    ContextScorer,
    RetrievalKRecallScorer,
)


@dataclass
class Scores:
    """Scores that are calculated by RagScoresCalculator.batch_score.

    If an input was not used by RagScoresCalculator.score, then the corresponding
    attribute is a list of Nones.

    Fields
    ------
    scores_list: List[Scores]
        List of scores that were calculated for each question and answer.
    question_list: List[Optional[str]]
        List of questions in the batch.
    reference_answer_list: List[Optional[str]]
        List of reference answers in the batch.
    llm_answer_list: List[Optional[str]]
        List of answers generated by the RAG system for the questions in the batch.
    retrieved_context_list_list: List[Optional[List[str]]]
        List of retrieved contexts used by the RAG system to make answers to the
        questions in the batch.
    top_k_context_list_list: List[Optional[List[str]]]
        List of top k contexts that would be retrieved by the RAG system for the
        questions in the batch. There's an assumption that the retrieved context list is
        a subset of the top k context list.
    """

    answer_similarity_score_list: List[Optional[float]]
    retrieval_precision_list: List[Optional[float]]
    augmentation_precision_list: List[Optional[float]]
    augmentation_accuracy_list: List[Optional[float]]
    answer_consistency_list: List[Optional[float]]
    answer_consistency_binary_list: List[Optional[float]]
    retrieval_k_recall_list: List[Optional[float]]
    overall_score_list: List[Optional[float]]
    question_list: List[Optional[str]]
    reference_answer_list: List[Optional[str]]
    llm_answer_list: List[Optional[str]]
    retrieved_context_list_list: List[Optional[List[str]]]
    top_k_context_list_list: List[Optional[List[str]]]

    def to_dataframe(self) -> pd.DataFrame:
        """Convert Scores to a pandas DataFrame."""
        df = pd.DataFrame(
            {
                "question": [x for x in self.question_list],
                "reference_answer": [x for x in self.reference_answer_list],
                "llm_answer": [x for x in self.llm_answer_list],
                "retrieved_context": [x for x in self.retrieved_context_list_list],
                "top_k_context": [x for x in self.top_k_context_list_list],
                "answer_similarity_score": [
                    x for x in self.answer_similarity_score_list
                ],
                "retrieval_precision": [x for x in self.retrieval_precision_list],
                "augmentation_precision": [x for x in self.augmentation_precision_list],
                "augmentation_accuracy": [x for x in self.augmentation_accuracy_list],
                "answer_consistency": [x for x in self.answer_consistency_list],
                "answer_consistency_binary": [
                    x for x in self.answer_consistency_binary_list
                ],
                "retrieval_k_recall": [x for x in self.retrieval_k_recall_list],
                "overall_score": [x for x in self.overall_score_list],
            }
        )
        return df.dropna(axis=1, how="all")

    def mean_scores(self) -> Dict[str, Optional[float]]:
        """Calculate the mean of the non-None scores.

        Returns
        -------
        Dict[str, Union[float, None]]
            Dictionary mapping each score name to its mean score.
        """
        if self.answer_similarity_score_list[0] is not None:
            mean_answer_similarity_score = mean(  # type: ignore
                self.answer_similarity_score_list
            )
        else:
            mean_answer_similarity_score = None
        if self.retrieval_precision_list[0] is not None:
            mean_retrieval_precision = mean(  # type: ignore
                self.retrieval_precision_list
            )
        else:
            mean_retrieval_precision = None
        if self.augmentation_precision_list[0] is not None:
            mean_augmentation_precision = mean(  # type: ignore
                self.augmentation_precision_list
            )
        else:
            mean_augmentation_precision = None
        if self.augmentation_accuracy_list[0] is not None:
            mean_augmentation_accuracy = mean(  # type: ignore
                self.augmentation_accuracy_list
            )
        else:
            mean_augmentation_accuracy = None
        if self.answer_consistency_binary_list[0] is not None:
            mean_answer_consistency_binary = mean(  # type: ignore
                self.answer_consistency_binary_list
            )
        else:
            mean_answer_consistency_binary = None
        if self.answer_consistency_list[0] is not None:
            mean_answer_consistency = mean(self.answer_consistency_list)  # type: ignore
        else:
            mean_answer_consistency = None
        if self.retrieval_k_recall_list[0] is not None:
            mean_retrieval_k_recall = mean(self.retrieval_k_recall_list)  # type: ignore
        else:
            mean_retrieval_k_recall = None
        if self.overall_score_list[0] is not None:
            mean_overall_score = mean(self.overall_score_list)  # type: ignore
        else:
            mean_overall_score = None
        return dict(
            answer_similarity_score=mean_answer_similarity_score,
            retrieval_precision=mean_retrieval_precision,
            augmentation_precision=mean_augmentation_precision,
            augmentation_accuracy=mean_augmentation_accuracy,
            answer_consistency_binary=mean_answer_consistency_binary,
            answer_consistency=mean_answer_consistency,
            retrieval_k_recall=mean_retrieval_k_recall,
            overall_score=mean_overall_score,
        )

    def scores_to_dict(self) -> Dict[str, List[Optional[float]]]:
        return dict(
            answer_similarity_score=self.answer_similarity_score_list,
            retrieval_precision=self.retrieval_precision_list,
            augmentation_precision=self.augmentation_precision_list,
            augmentation_accuracy=self.augmentation_accuracy_list,
            answer_consistency_binary=self.answer_consistency_binary_list,
            answer_consistency=self.answer_consistency_list,
            retrieval_k_recall=self.retrieval_k_recall_list,
            overall_score=self.overall_score_list,
        )


class RagScoresCalculator(object):
    """Class for calculating any collection of RAG scores.

    Note: If all scores are set to False, then the default is to calculate the answer
    similarity score, retrieval precision, augmentation precision, augmentation
    accuracy, and answer consistency.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    answer_similarity_score: bool
        Whether to calculate the answer similarity score.
    retrieval_precision: bool
        Whether to calculate the retrieval precision score.
    augmentation_precision: bool
        Whether to calculate the augmentation precision score.
    augmentation_accuracy: bool
        Whether to calculate the augmentation accuracy score.
    answer_consistency: bool
        Whether to calculate the answer consistency score.
    answer_consistency_binary: bool
        Whether to calculate the answer consistency binary score.
    retrieval_k_recall: bool
        Whether to calculate the retrieval k-recall score.
    """

    def __init__(
        self,
        model: str,
        answer_similarity_score: bool = False,
        retrieval_precision: bool = False,
        augmentation_precision: bool = False,
        augmentation_accuracy: bool = False,
        answer_consistency: bool = False,
        answer_consistency_binary: bool = False,
        retrieval_k_recall: bool = False,
    ):
        # in this case we do the defualt because the user has not specified any scores
        if not (
            answer_similarity_score
            or retrieval_precision
            or augmentation_precision
            or augmentation_accuracy
            or answer_consistency
            or answer_consistency_binary
            or retrieval_k_recall
        ):
            answer_similarity_score = True
            retrieval_precision = True
            augmentation_precision = True
            augmentation_accuracy = True
            answer_consistency = True
        self.model = model
        self.answer_similarity_score = answer_similarity_score
        self.retrieval_precision = retrieval_precision
        self.augmentation_precision = augmentation_precision
        self.augmentation_accuracy = augmentation_accuracy
        self.answer_consistency_binary = answer_consistency_binary
        self.answer_consistency = answer_consistency
        self.retrieval_k_recall = retrieval_k_recall

        if answer_similarity_score:
            self.answer_similarity_scorer = AnswerSimilarityScorer(self.model)
        else:
            self.answer_similarity_scorer = None  # type: ignore

        if retrieval_precision and augmentation_precision and augmentation_accuracy:
            self.context_scorer = ContextScorer(self.model)
        elif not (
            retrieval_precision and augmentation_precision and augmentation_accuracy
        ):
            self.context_scorer = None  # type: ignore
        else:
            error_message = (
                "must have all three of retrieval_precision, augmentation_precision, "
                "and augmentation_accuracy set to True or False"
            )
            raise ValueError(error_message)

        if answer_consistency_binary:
            self.answer_consistency_binary_scorer = AnswerConsistencyBinaryScorer(
                self.model
            )
        else:
            self.answer_consistency_binary_scorer = None  # type: ignore

        if answer_consistency:
            self.answer_consistency_scorer = AnswerConsistencyScorer(self.model)
        else:
            self.answer_consistency_scorer = None  # type: ignore
        if retrieval_k_recall:
            self.retrieval_k_recall_scorer = RetrievalKRecallScorer(self.model)
        else:
            self.retrieval_k_recall_scorer = None  # type: ignore

    def validate_score_parameters(
        self,
        question: Optional[str] = None,
        reference_answer: Optional[str] = None,
        llm_answer: Optional[str] = None,
        retrieved_context_list: Optional[List[str]] = None,
        top_k_context_list: Optional[List[str]] = None,
    ) -> None:
        if self.answer_similarity_scorer is not None:
            if question is None or reference_answer is None or llm_answer is None:
                error_message = (
                    "Need question, reference_answer, and llm_answer for similarity "
                    "scorer"
                )
                raise ValueError(error_message)
        if self.context_scorer is not None:
            if question is None or llm_answer is None or retrieved_context_list is None:
                error_message = (
                    "Need question, llm_answer, and retrieved_context_list for "
                    "context scores scorer"
                )
                raise ValueError(error_message)
        if self.answer_consistency_binary_scorer is not None:
            if llm_answer is None or retrieved_context_list is None:
                error_message = (
                    "Need llm_answer and retrieved_context_list for answer consistency "
                    "binary scorer"
                )
                raise ValueError(error_message)
        if self.answer_consistency_scorer is not None:
            if llm_answer is None or retrieved_context_list is None:
                error_message = (
                    "Need llm_answer and retrieved_context_list for answer consistency "
                    "scorer"
                )
                raise ValueError(error_message)
        if self.retrieval_k_recall_scorer is not None:
            if (
                question is None
                or retrieved_context_list is None
                or top_k_context_list is None
            ):
                error_message = (
                    "Need question, retrieved_context_list, and top_k_context_list for "
                    "retrieval k recall scorer"
                )
                raise ValueError(error_message)

    def calculate_overall_score(
        self,
        answer_similarity_score: Optional[float],
        retrieval_precision: Optional[float],
        augmentation_precision: Optional[float],
        augmentation_accuracy: Optional[float],
        answer_consistency_binary: Optional[float],
        answer_consistency: Optional[float],
        retrieval_k_recall: Optional[float],
    ) -> float:
        score_sum = 0.0
        score_count = 0
        if answer_similarity_score is not None:
            score_sum += answer_similarity_score / 5
            score_count += 1
        if retrieval_precision is not None:
            score_sum += retrieval_precision
            score_count += 1
        if augmentation_precision is not None:
            score_sum += augmentation_precision
            score_count += 1
        if augmentation_accuracy is not None:
            score_sum += augmentation_accuracy
            score_count += 1
        if answer_consistency_binary is not None:
            score_sum += answer_consistency_binary
            score_count += 1
        if answer_consistency is not None:
            score_sum += answer_consistency
            score_count += 1
        if retrieval_k_recall is not None:
            score_sum += retrieval_k_recall
            score_count += 1
        return score_sum / score_count

    def score(
        self,
        question: Optional[str] = None,
        reference_answer: Optional[str] = None,
        llm_answer: Optional[str] = None,
        retrieved_context_list: Optional[List[str]] = None,
        top_k_context_list: Optional[List[str]] = None,
    ) -> Scores:
        """Calculate RAG scores.

        Note: Only the inputs needed to calculate the scores set in __init__ are
        necessary.

        Parameters
        ----------
        question: Optional[str]
            The question that was asked.
        reference_answer: Optional[str]
            The answer that was expected.
        llm_answer: Optional[str]
            The answer that was generated by the RAG system.
        retrieved_context_list: Optional[List[str]]
            Retrieved context used by the RAG system to make answer.
        top_k_context_list: Optional[List[str]]
            Top k contexts that would be retrieved by the RAG system. There's an
            assumption that the retrieved context list is a subset of the top k context
            list.

        Returns
        -------
        Scores
            All the scores that were calculated.
        """
        self.validate_score_parameters(
            question,
            reference_answer,
            llm_answer,
            retrieved_context_list,
            top_k_context_list,
        )
        if self.answer_similarity_scorer is not None:
            answer_similarity_score = self.answer_similarity_scorer.score(
                question, reference_answer, llm_answer  # type: ignore
            )
        else:
            answer_similarity_score = None

        if self.context_scorer is not None:
            context_scores = self.context_scorer.score(
                question, llm_answer, retrieved_context_list  # type: ignore
            )
            retrieval_precision = context_scores.retrieval_precision
            augmentation_precision = context_scores.augmentation_precision
            augmentation_accuracy = context_scores.augmentation_accuracy
        else:
            retrieval_precision = None
            augmentation_precision = None
            augmentation_accuracy = None

        if self.answer_consistency_binary_scorer is not None:
            answer_consistency_binary = self.answer_consistency_binary_scorer.score(
                llm_answer, retrieved_context_list  # type: ignore
            )
        else:
            answer_consistency_binary = None

        if self.answer_consistency_scorer is not None:
            answer_consistency_score = self.answer_consistency_scorer.score(
                llm_answer, retrieved_context_list  # type: ignore
            )
            annswer_consistency = answer_consistency_score.score
        else:
            annswer_consistency = None

        if self.retrieval_k_recall_scorer is not None:
            if top_k_context_list is None:
                raise ValueError(
                    "top_k_context_list must be provided if retrieval_k_recall is True"
                )
            retrieval_k_recall_score = self.retrieval_k_recall_scorer.score(
                question, retrieved_context_list, top_k_context_list  # type: ignore
            )
            retrieval_k_recall = retrieval_k_recall_score.score
        else:
            retrieval_k_recall = None

        overall_score = self.calculate_overall_score(
            answer_similarity_score,
            retrieval_precision,
            augmentation_precision,
            augmentation_accuracy,
            answer_consistency_binary,
            annswer_consistency,
            retrieval_k_recall,
        )

        return Scores(
            answer_similarity_score_list=[answer_similarity_score],
            retrieval_precision_list=[retrieval_precision],
            augmentation_precision_list=[augmentation_precision],
            augmentation_accuracy_list=[augmentation_accuracy],
            answer_consistency_binary_list=[answer_consistency_binary],
            answer_consistency_list=[annswer_consistency],
            retrieval_k_recall_list=[retrieval_k_recall],
            overall_score_list=[overall_score],
            question_list=[question],
            reference_answer_list=[reference_answer],
            llm_answer_list=[llm_answer],
            retrieved_context_list_list=[retrieved_context_list],
            top_k_context_list_list=[top_k_context_list],
        )

    def get_n_samples(
        self,
        question_list: Optional[List[Optional[str]]],
        retrieved_context_list_list: Optional[List[Optional[List[str]]]],
    ) -> int:
        if question_list is not None:
            n_samples = len(question_list)
        elif retrieved_context_list_list is not None:
            n_samples = len(retrieved_context_list_list)
        else:
            error_message = (
                "Need either question_list or retrieved_context_list_list to calculate "
                "metrics. Got neither."
            )
            raise ValueError(error_message)
        return n_samples

    def score_batch(
        self,
        question_list: Optional[List[Optional[str]]] = None,
        reference_answer_list: Optional[List[Optional[str]]] = None,
        llm_answer_list: Optional[List[Optional[str]]] = None,
        retrieved_context_list_list: Optional[List[Optional[List[str]]]] = None,
        top_k_context_list_list: Optional[List[Optional[List[str]]]] = None,
    ) -> Scores:
        """Calculate RAG scores for a batch of questions and answers.

        Notes:
            - Only the inputs needed to calculate the scores set in __init__ are
              necessary.
            - All inputs are lists that are assumed to be the same size with
              corresponding elements in the same positions.

        Parameters
        ----------
        question_list: Optional[List[Optional[str]]]
            Batch of questions that were asked.
        reference_answer_list: Optional[List[Optional[str]]]
            Batch of answers that were expected.
        llm_answer_list: Optional[List[Optional[str]]]
            Batch of answers generated by the RAG system.
        retrieved_context_list_list: Optional[List[Optional[List[str]]]]
            Batch of retrieved contexts used by the RAG system to make answer.
        top_k_context_list_list: Optional[List[Optional[List[str]]]]
            Batch of top k contexts that would be retrieved by the RAG system. There's
            an assumption that the retrieved context list is a subset of the top k
            context list.

        Returns
        -------
        Scores
            All the scores that were calculated.
        """
        n_samples = self.get_n_samples(question_list, retrieved_context_list_list)
        if question_list is None:
            question_list = [None for _ in range(n_samples)]
        if reference_answer_list is None:
            reference_answer_list = [None for _ in range(n_samples)]
        if llm_answer_list is None:
            llm_answer_list = [None for _ in range(n_samples)]
        if retrieved_context_list_list is None:
            retrieved_context_list_list = [None for _ in range(n_samples)]
        if top_k_context_list_list is None:
            top_k_context_list_list = [None for _ in range(n_samples)]

        score_list = []
        for (
            question,
            reference_answer,
            llm_answer,
            retrieved_context_list,
            top_k_context_list,
        ) in zip(
            question_list,
            reference_answer_list,
            llm_answer_list,
            retrieved_context_list_list,
            top_k_context_list_list,
        ):
            score = self.score(
                question,
                reference_answer,
                llm_answer,
                retrieved_context_list,
                top_k_context_list,
            )
            score_list.append(score)

        return Scores(
            answer_similarity_score_list=[
                x.answer_similarity_score_list[0] for x in score_list
            ],
            retrieval_precision_list=[
                x.retrieval_precision_list[0] for x in score_list
            ],
            augmentation_precision_list=[
                x.augmentation_precision_list[0] for x in score_list
            ],
            augmentation_accuracy_list=[
                x.augmentation_accuracy_list[0] for x in score_list
            ],
            answer_consistency_binary_list=[
                x.answer_consistency_binary_list[0] for x in score_list
            ],
            answer_consistency_list=[x.answer_consistency_list[0] for x in score_list],
            retrieval_k_recall_list=[x.retrieval_k_recall_list[0] for x in score_list],
            overall_score_list=[x.overall_score_list[0] for x in score_list],
            question_list=question_list,
            reference_answer_list=reference_answer_list,
            llm_answer_list=llm_answer_list,
            retrieved_context_list_list=retrieved_context_list_list,
            top_k_context_list_list=top_k_context_list_list,
        )



import logging
from typing import List

from tvalmetrics.openai_api_util import get_single_message

logger = logging.getLogger()

NEWLINE_INDENT_STRING = "\n    "


def ask_for_similarity_score(
    question: str, reference_answer: str, llm_answer: str, model: str
) -> str:
    """Sends prompt for answer similarity score to OpenAI API, and returns response.

    Parameters
    ----------
    question: str
        The question that was asked.
    reference_answer: str
        The answer that was expected.
    llm_answer: str
        The answer that was generated by the RAG system.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    logger.debug(f"Asking {model} for similarity score for question: {question}")
    main_message = """
    Considering the reference answer and the new answer to the following question, on
    a scale of 0 to 5, where 5 means the same and 0 means not similar, how similar in
    meaning is the new answer to the reference answer? Respond with just a number and
    no additional text.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    main_message += f"\nQUESTION: {question}\n"
    main_message += f"REFERENCE ANSWER: {reference_answer}\n"
    main_message += f"NEW ANSWER: {llm_answer}\n"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message


def ask_whether_answer_is_consistent_with_context(
    answer: str, context_list: List[str], model: str
) -> str:
    """Sends prompt for answer consistency binary score and returns response.

    Parameters
    ----------
    answer: str
        The answer that was generated by the RAG system.
    context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """

    logger.debug(f"Asking {model} whether answer hallucinates")
    main_message = """
    Considering the following list of context and then answer, which answers a
    user's query using the context, determine whether the answer contains any
    information that can not be attributed to the intormation in the list of
    context. If the answer contains information that cannot be attributed to the
    context then respond with true. Otherwise response with false. Response with
    either true or false and no additional text.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    for i, context in enumerate(context_list):
        main_message += f"\n\nCONTEXT {i}:\n{context}\nEND OF CONTEXT {i}"
    main_message += f"\n\nANSWER: {answer}"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message


def ask_whether_context_is_relevant(question: str, context: str, model: str) -> str:
    """Sends prompt to get context relevance to Open AI API and returns response.

    Parameters
    ----------
    question: str
        The question that was asked.
    context: str
        One piece of context retrieved by RAG system.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    logger.debug(f"Asking {model} for context relevance for question {question}")
    main_message = """
    Considering the following question and context, determine whether the context
    is relevant for answering the question. If the context is relevant for answering
    the question, responsd with true. If the context is not relevant for answering
    the question, respond with false. Respond with either true or false and no
    additinal text.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    main_message += f"\nQUESTION: {question}\n"
    main_message += f"CONTEXT: {context}\n"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message


def ask_whether_answer_contains_context(answer: str, context: str, model: str) -> str:
    """Sends prompt for whether answer contains context and returns response.

    Parameters
    ----------
    answer: str
        The answer that was generated by the RAG system.
    context: str
        One piece of context retrieved by RAG system.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    logger.debug(f"Asking {model} whether answer contains context")
    main_message = """
    Considering the following answer and context, determine whether the answer
    contains information derived from the context. If the answer contains
    information derived from the context, respond with true. If the answer does not
    contain information derived from the context, respond with false.  Respond with
    either true or false and no additinal text.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    main_message += f"\nANSWER: {answer}\n"
    main_message += f"CONTEXT: {context}\n"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message


def ask_for_main_points(answer: str, model: str) -> str:
    """Sends prompt for main points in answer to Open AI API and returns response.

    Parameters
    ----------
    answer: str
        The answer that was generated by the RAG system.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    logger.debug(f"Asking {model} for bullet list of main points in answer")
    main_message = """
    Write down in a bulleted list using markdown (so each bullet is a "*"), the main
    points in the following answer to a user's query. Respond with the bulleted list
    and no additional text. Only use a single "*" for each bullet and do not use a "*"
    anywhere in your response except for the bullets.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    main_message += f"\nANSWER: {answer}"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message


def ask_whether_statement_derived_from_context(
    statement: str, context_list: List[str], model: str
) -> str:
    """Sends prompt for whether statement is derived from context and returns response.

    Parameters
    ----------
    statement: str
        The statement to be checked.
    context_list: List[str]
        List of retrieved context to see if statement is derived from this context.
    model: str
        Name of the LLM model to use as the LLM evaluator.

    Returns
    -------
    str
        Response from OpenAI API.
    """
    logger.debug(f"Asking {model} whether statement is derived from context")
    main_message = """
    Considering the following statement and then list of context, determine whether the
    statement can be derived from the context. If the statement can be derived from the
    context response with true. Otherwise response with false. Respond with either true
    or false and no additional text.
    """.replace(
        NEWLINE_INDENT_STRING, " "
    ).strip()
    main_message += f"\n\nSTATEMENT:\n{statement}\nEND OF STATEMENT"
    for i, context in enumerate(context_list):
        main_message += f"\n\nCONTEXT {i}:\n{context}\nEND OF CONTEXT {i}"

    response_message = get_single_message(main_message, model, temperature=0.0)

    return response_message



import logging

from tvalmetrics.llm_calls import ask_for_similarity_score

logger = logging.getLogger()


class AnswerSimilarityScorer(object):
    """Class for calculating answer similarity score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(self, question: str, reference_answer: str, llm_answer: str) -> float:
        """Calculate answer similarity score.

        Parameters
        ----------
        question: str
            The question that was asked.
        reference_answer: str
            The answer that was expected.
        llm_answer: str
            The answer that was generated by the RAG system.

        Returns
        -------
        float
            Float between 0 and 5 representing the answer similarity score.
        """
        similarity_score_response = ask_for_similarity_score(
            question, reference_answer, llm_answer, self.model
        )
        try:
            similarity_score = float(similarity_score_response)
        except ValueError:
            error_message = (
                f"Failed to parse similarity score {similarity_score} as "
                "float, setting score to 0.0"
            )
            logger.error(error_message)
            similarity_score = 0.0

        return similarity_score



from tvalmetrics.scorers.answer_consistency_binary_scorer import (  # noqa F401
    AnswerConsistencyBinaryScorer,
)
from tvalmetrics.scorers.answer_consistency_scorer import (  # noqa F401
    AnswerConsistencyScorer,
)
from tvalmetrics.scorers.answer_similarity_scorer import (  # noqa F401
    AnswerSimilarityScorer,
)
from tvalmetrics.scorers.augmentation_accuracy_scorer import (  # noqa F401
    AugmentationAccuracyScorer,
)
from tvalmetrics.scorers.augmentation_precision_scorer import (  # noqa F401
    AugmentationPrecisionScorer,
)
from tvalmetrics.scorers.context_scorer import ContextScorer  # noqa F401
from tvalmetrics.scorers.retrieval_k_recall_scorer import (  # noqa F401
    RetrievalKRecallScorer,
)
from tvalmetrics.scorers.retrieval_precision_scorer import (  # noqa F401
    RetrievalPrecisionScorer,
)



from dataclasses import dataclass
from typing import List

from tvalmetrics.llm_calls import (
    ask_for_main_points,
    ask_whether_statement_derived_from_context,
)
from tvalmetrics.scorers.scorers_util import (
    parse_boolean_response,
    parse_bullet_list_response,
)


@dataclass
class AnswerConsistency:
    """Answer consistency score and information used to calculate the score.

    Fields
    ------
    score: float
        Float between 0 and 1 representing the answer consistency score.
    answer: str
        The answer that was generated by the RAG system.
    context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    main_point_list: List[str]
        List of main points of answer generated by the LLM evaluator.
    main_point_derived_from_context_list: List[bool]
        List of booleans representing whether each main point in main_point_list was
        derived from context in context_list.
    """

    score: float
    answer: str
    context_list: List[str]
    main_point_list: List[str]
    main_point_derived_from_context_list: List[bool]


class AnswerConsistencyScorer(object):
    """Class for calculating answer consistency score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(self, answer: str, context_list: List[str]) -> AnswerConsistency:
        """Calculate answer consistency score.

        Parameters
        ----------
        answer: str
            The answer that was generated by the RAG system.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        AnswerConsistencyScore
            The score and information used to calculate the score.
        """
        main_points_response = ask_for_main_points(answer, self.model)
        main_point_list = parse_bullet_list_response(main_points_response)
        main_point_derived_from_context_list = []
        for main_point in main_point_list:
            statement_derived_from_context_response = (
                ask_whether_statement_derived_from_context(
                    main_point, context_list, self.model
                )
            )
            main_point_derived_from_context_list.append(
                parse_boolean_response(statement_derived_from_context_response)
            )
        score = sum(main_point_derived_from_context_list) / len(main_point_list)
        return AnswerConsistency(
            score,
            answer,
            context_list,
            main_point_list,
            main_point_derived_from_context_list,
        )



from typing import List

from tvalmetrics.llm_calls import ask_whether_answer_is_consistent_with_context
from tvalmetrics.scorers.scorers_util import parse_boolean_response


class AnswerConsistencyBinaryScorer(object):
    """Class for calculating answer consistency binary score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(self, answer: str, context_list: List[str]) -> int:
        """Calculate answer consistency binary score.

        Parameters
        ----------
        answer: str
            The answer to be scored.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        int
            0 if there is information in answer not derived from context in context_list
            1 otherwise
        """
        hallucination_response = ask_whether_answer_is_consistent_with_context(
            answer, context_list, self.model
        )
        hallucination = parse_boolean_response(hallucination_response)
        return int(hallucination)



from dataclasses import dataclass
from typing import List

from tvalmetrics.llm_calls import ask_whether_context_is_relevant
from tvalmetrics.scorers.scorers_util import parse_boolean_response


@dataclass
class RetrievalPrecision:
    """Retrieval precision score and information used to calculate the score.

    Fields
    ------
    score: float
        Float between 0 and 1 representing the retrieval precision score.
    question: str
        The question that was asked.
    context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    context_relevant_list: List[bool]
        List of booleans representing whether each context in context_list is relevant
        for answering the question.
    """

    score: float
    question: str
    context_list: List[str]
    context_relevant_list: List[bool]


class RetrievalPrecisionScorer(object):
    """Class for calculating retrieval precision score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(self, question: str, context_list: List[str]) -> RetrievalPrecision:
        """Calculate retrieval precision score.

        Parameters
        ----------
        question: str
            The question that was asked.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        RetrievalPrecisionScore
            Retrieval precision score and information used to calculate the score."""
        context_relevant_list = []
        for context in context_list:
            relevance_response = ask_whether_context_is_relevant(
                question, context, self.model
            )
            context_relevant_list.append(parse_boolean_response(relevance_response))

        score = sum(context_relevant_list) / len(context_relevant_list)
        return RetrievalPrecision(score, question, context_list, context_relevant_list)



from dataclasses import dataclass
from typing import List

from tvalmetrics.llm_calls import ask_whether_answer_contains_context
from tvalmetrics.scorers.scorers_util import parse_boolean_response


@dataclass
class AugmentationAccuracy:
    """Augmentation accuracy score and information used to calculate the score.

    Fields
    ------
    score: float
        Float between 0 and 1 representing the augmentation accuracy score.
    answer: str
        The answer that was generated by the RAG system.
    context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    answer_contains_context_list: List[bool]
        List of booleans representing whether information from each context in
        context_list is in answer.
    """

    score: float
    answer: str
    context_list: List[str]
    answer_contains_context_list: List[bool]


class AugmentationAccuracyScorer(object):
    """Class for calculating augmentation accuracy score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(self, answer: str, context_list: List[str]) -> AugmentationAccuracy:
        """Calculate augmentation accuracy score.

        Parameters
        ----------
        answer: str
            The answer that was generated by the RAG system.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        AugmentationAccuracyScore
            Augmentation accuracy score and information used to calculate the score.
        """
        contains_context_list = []
        for context in context_list:
            contains_context_response = ask_whether_answer_contains_context(
                answer, context, self.model
            )
            contains_context_list.append(
                parse_boolean_response(contains_context_response)
            )

        score = sum(contains_context_list) / len(contains_context_list)
        return AugmentationAccuracy(score, answer, context_list, contains_context_list)



from dataclasses import dataclass
from typing import List

from tvalmetrics.scorers.augmentation_accuracy_scorer import AugmentationAccuracyScorer
from tvalmetrics.scorers.augmentation_precision_scorer import (
    AugmentationPrecisionScorer,
)
from tvalmetrics.scorers.retrieval_precision_scorer import RetrievalPrecisionScorer


@dataclass
class ContextScores:
    """Retrieval precision, augmentation precision, and augmentation accuracy scores.

    Fields
    ------
    retrieval_precision: float
        Float between 0 and 1 representing the retrieval precision score.
    augmentation_precision: float
        Float between 0 and 1 representing the augmentation precision score.
    augmentation_accuracy: float
        Float between 0 and 1 representing the augmentation accuracy score."""

    retrieval_precision: float
    augmentation_precision: float
    augmentation_accuracy: float


class ContextScorer(object):
    """Class for calculating the three main context scores at the same time.

    Calculates retrieval precision, augmentation precision, and augmentation accuracy at
    the same time to avoid repeating the same LLM calls that would happen if each score
    was calculated separately.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model
        self.retrieval_precision_scorer = RetrievalPrecisionScorer(self.model)
        self.augmentation_accuracy_scorer = AugmentationAccuracyScorer(self.model)

    def score(
        self, question: str, answer: str, context_list: List[str]
    ) -> ContextScores:
        """Calculate the three main context scores.

        Caclulates retrieval precision, augmentation precision, and augmentation
        accuracy.

        Parameters
        ----------
        question: str
            The question that was asked.
        answer: str
            The answer that was generated by the RAG system.
        context_list: List[str]
            Retrieved context used by the RAG system to make answer.

        Returns
        -------
        ContextScores
            Retrieval precision, augmentation precision, and augmentation accuracy
            scores.
        """
        retrieval_precision_score = self.retrieval_precision_scorer.score(
            question, context_list
        )
        retrieval_precision = retrieval_precision_score.score
        context_relevant_list = retrieval_precision_score.context_relevant_list
        augmentation_accuracy_score = self.augmentation_accuracy_scorer.score(
            answer, context_list
        )
        augmentation_accuracy = augmentation_accuracy_score.score
        contains_context_list = augmentation_accuracy_score.answer_contains_context_list
        augmentation_precision = AugmentationPrecisionScorer.score_from_context_labels(
            context_relevant_list, contains_context_list
        )

        return ContextScores(
            retrieval_precision=retrieval_precision,
            augmentation_precision=augmentation_precision,
            augmentation_accuracy=augmentation_accuracy,
        )



from dataclasses import dataclass
from typing import List

from tvalmetrics.llm_calls import ask_whether_context_is_relevant
from tvalmetrics.scorers.scorers_util import parse_boolean_response


@dataclass
class RetrievalKRecall:
    """Retrieval k-recall score and information used to calculate the score.

    Fields
    ------
    score: float
        Retrieval k-recall score.
    k: int
        Number of total contexts retrieved.
    retrieved_context_list: List[str]
        Retrieved context used by the RAG system to make answer.
    top_k_context_list: List[str]
        Top k context retrieved by the RAG system. retrieved_context_list is a subset of
        top_k_context_list.
    top_k_context_relevant_list: List[bool]
        List of booleans indicating whether the context is relevant or not to answering
        the question.
    """

    score: float
    k: int
    retrieved_context_list: List[str]
    top_k_context_list: List[str]
    top_k_context_relevant_list: List[bool]


class RetrievalKRecallScorer(object):
    """Class for calculating retrieval k-recall score.

    Parameters
    ----------
    model: str
        Name of the LLM model to use as the LLM evaluator.
    """

    def __init__(self, model: str):
        self.model = model

    def score(
        self,
        question: str,
        retrieved_context_list: List[str],
        top_k_context_list: List[str],
    ) -> RetrievalKRecall:
        """Calculate retrieval k-recall score.

        Parameters
        ----------
        question: str
            The question that was asked.
        retrieved_context_list: List[str]
            Retrieved context used by the RAG system to make answer.
        top_k_context_list: List[str]
            Top k contexts that would be retrieved by the RAG system. There's an
            assumption that the retrieved context list is a subset of the top k context
            list.

        Returns
        -------
        RetrievalKRecallScore
            Retrieval k-recall score and information used to calculate the score.
        """
        k = len(top_k_context_list)
        top_k_context_relevant_list = []
        retrieved_context_relevant_cnt = 0
        for context in top_k_context_list:
            relevance_response = ask_whether_context_is_relevant(
                question, context, self.model
            )
            relevant = parse_boolean_response(relevance_response)
            top_k_context_relevant_list.append(relevant)
            if relevant:
                if context in retrieved_context_list:
                    retrieved_context_relevant_cnt += 1
        if len(top_k_context_relevant_list) == 0:
            score = 0.0
        else:
            score = retrieved_context_relevant_cnt / sum(top_k_context_relevant_list)

        return RetrievalKRecall(
            score,
            k,
            retrieved_context_list,
            top_k_context_list,
            top_k_context_relevant_list,
        )
